/**************************************************************************\
 gatos (General ATI TV and Overlay Software)

  Project Coordinated By Insomnia (Steaphan Greene)
  (insomnia@core.binghamton.edu)

  Copyright (C) 1999 Steaphan Greene, yvind Aabling, Octavian Purdila, 
	Vladimir Dergachev, Christian Lupien and Chris Hardy

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

\**************************************************************************/

#define GATOS_TVOUT_C 1

#include "gatos.h"
#include "tvout.h"
#include "m64regs.h"
#include "r128regs.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <math.h>

/* MPP_CONFIG values */
#define MPPNORMAL	0x80038CB0L	/* write */
#define MPPREAD		0x84038CB0L	/* prefetch read */
#define MPPNORMALINC	0x80338CB0L	/* write, autoincrement MPP_ADDR */
#define MPPREADINC	0x84338CB0L	/* prefetch read, MPP_ADDR autoincr */
/* MPP_ADDR autoincrement wraps around after 4 adresses (2 LSB's only) */

#define MPPTRIES	100000	/* See comments in mpp_wait() /AA */

/* Read a byte with possible bugfix correction from MPP */
#define MPPREADBYTE	(MPP_DATA0 ^ gatos.bugfix_mppread)

/* Private global vars */
static char *ident=NULL, *name=NULL ;
static int id=0, rev=0 ;
static u32 timeouts=0 ;

/* Note:
 * ImpacTV registers are 32 bit, indexed by the 16 bit
 * value in MPP registers 0x0009 (MSB) and 0x0008 (LSB).
 * Value of indexed ImpacTV register is accessed
 * through MPP registers 0x0018 (LSB) to 0x001B (MSB).
 * tvout_read_i2c_cntl8() and tvout_write_i2c_cntl8() assumes that
 * the ImpacTV Register Index points to TV_I2C_CNTL (index 0x0015)
 * and that MPP_ADDR points to low byte of ImpacTV Register Data.
 * All other routines must call tvout_addr(TV_I2C_CNTL) to restore
 * these conditions after accessing other ImpacTV chip registers.
 * AA
 */

/* ------------------------------------------------------------------------ */
/* Initialization routine */

int tvout_init(void) {

  u8 init ; u32 save1, save2, save3 ;

  /* The Mach64 TV Out driver hasn't been ported to the Rage128 yet /AA */
  if (ati.r128) RETURN0 ;

  /* ATI chip with MPP read bug ? */
  if ( ati.gt_c2u1 || ati.gt_c2u2 )
    if ( ati.gp_c2u1 || ati.gq_c2u1 ||
         ati.gp_c2u2 || ati.gq_c2u2 ) {
    if (VERBOSE) {
      printf("TV Out: Chip with MPP read bug (correcting)\n") ;
      fflush(stdout) ; }
    gatos.bugfix_mppread = 0x81 ; }

  save1 = DAC_CNTL ; save2 = MPP_CONFIG ; save3 = MPP_STROBE_SEQ ;
  /* strobe needs to be 0x8383 on AIW otherwise the bus stays addressed
     and prevents the impacTV from syncing, CL */
  /* reverted to 1.13 version, Jon
  MPP_STROBE_SEQ = 0x00008387L ; DAC_CNTL &= 0xFFFFBFFFL ;
  */
  MPP_STROBE_SEQ = 0x00000383L ; DAC_CNTL &= 0xFFFFBFFFL ;


  /* Read ImpacTV Revision Code and Initialize value */
  MPP_CONFIG = MPPREAD ; mpp_wait("tvout_init()") ;
  MPP_ADDR = 0x0000000AL ; init = MPPREADBYTE ; mpp_wait("tvout_init() 1") ;
  MPP_ADDR = 0x0000000BL ; id   = MPPREADBYTE ; mpp_wait("tvout_init() 2") ;
  MPP_CONFIG = MPPNORMAL ;

  switch (id) {
    case IMPACTV:                    ident = "ImpacTV" ;   rev = 1 ; break ;
    case IMPACTV2:  case IMPACTV2M:  ident = "ImpacTV2" ;  rev = 2 ; break ;
    case IMPACTV2P: case IMPACTV2PM: ident = "ImpacTV2+" ; rev = 3 ; break ;
    default: break ; }
  switch (id) {
    case IMPACTV:    name = "ImpacTV" ; break ;
    case IMPACTV2:   name = "ImpacTV2 (Macrovision disabled)" ; break ;
    case IMPACTV2M:  name = "ImpacTV2 (Macrovision enabled)" ; break ;
    case IMPACTV2P:  name = "ImpacTV2+ (Macrovision disabled)" ; break ;
    case IMPACTV2PM: name = "ImpacTV2+ (Macrovision enabled)" ; break ;
    default: break ; }
  if (init != 0x00) rev = 0 ;

  if (rev) {
    if (VERBOSE) { printf("TV Out: %s\n",name) ; fflush(stdout) ; }
    tvout_addr(TV_I2C_CNTL) ; }
  else { /* Initialization failed, restore registers */
    if (VERBOSE) {
      printf("TV Out: No ImpacTV chip (ID read was %02X)\n", id) ;
      fflush(stdout) ; }
    DAC_CNTL = save1 ; MPP_CONFIG = save2 ; MPP_STROBE_SEQ = save3 ; }

  /* Always return success (even if ImpacTV chip not present :-) */
  RETURN0 ; }

/* ------------------------------------------------------------------------ */
/* Public routines */

/* For MPP/ImpacTV I2C Driver */
int tvout_present(void) { return (rev!=0) ; }

/* Read low byte of ImpacTV TV_I2C_CNTL register */
u8 tvout_read_i2c_cntl8(void) {
  mpp_wait("tvout_read_i2c_cntl8()") ;
  MPP_CONFIG = MPPREAD ; mpp_wait("tvout_read_i2c_cntl8() 1") ;
  return MPPREADBYTE ; }

/* Write low byte of ImpacTV TV_I2C_CNTL register */
void tvout_write_i2c_cntl8(u8 data) {
  mpp_wait("tvout_write_i2c_cntl8()") ;
  MPP_CONFIG = MPPNORMAL ; MPP_DATA0 = data ; }

/* Read ImpacTV register */
u32 tvout_read32(u16 addr) {
  u32 data ;
  tvout_addr(addr) ;
  MPP_CONFIG = MPPREADINC ; mpp_wait("tvout_read32()") ;
  data  = MPPREADBYTE ;       mpp_wait("tvout_read32() 1") ;
  data |= MPPREADBYTE << 8 ;  mpp_wait("tvout_read32() 2") ;
  data |= MPPREADBYTE << 16 ; mpp_wait("tvout_read32() 3") ;
  data |= MPPREADBYTE << 24 ; mpp_wait("tvout_read32() 4") ;
  if (addr != TV_I2C_CNTL) tvout_addr(TV_I2C_CNTL) ; return data ; }

/* Write to ImpacTV register */
void tvout_write32(u16 addr, u32 data) {
  tvout_addr(addr) ;
  mpp_wait("tvout_write32()") ;
  MPP_CONFIG = MPPNORMALINC ;
  MPP_DATA0 = data ;       mpp_wait("tvout_write32() 1") ;
  MPP_DATA0 = data >> 8 ;  mpp_wait("tvout_write32() 2") ;
  MPP_DATA0 = data >> 16 ; mpp_wait("tvout_write32() 3") ;
  MPP_DATA0 = data >> 24 ; mpp_wait("tvout_write32() 4") ;
  if (addr != TV_I2C_CNTL) tvout_addr(TV_I2C_CNTL) ; }

/* ------------------------------------------------------------------------ */
/* Private routines */

/* Set ImpacTV register index and return MPP_ADDR
 * to 0x0018 for ImpacTV register data access. */
static void tvout_addr(u16 addr) {
  mpp_wait("tvout_addr()") ;
  MPP_CONFIG = MPPNORMALINC ; MPP_ADDR = 0x00000008L ;
  MPP_DATA0 = addr ;    mpp_wait("tvout_addr() 1") ;
  MPP_DATA0 = addr>>8 ; mpp_wait("tvout_addr() 2") ;
  MPP_CONFIG = MPPNORMAL ; MPP_ADDR = 0x00000018L ; }

/* Returns 1 if ready, 0 if MPP bus timed out (not ready) */
static int mpp_wait(char *str) {
  u32 tries=MPPTRIES ;
  /* For some reason, it's *necessary* to have *two* while() poll loops
   * to avoid MPP bus timeouts, but doubling MPPTRIES doesn't help ...
   * Maybe the code optimizer is playing tricks on us ?!?
   * Nah, can't be, problem is there with 'gcc -g' as well ...
   * Also, it seems to make no difference if MPPTRIES is 0x100
   * (the official ATI value) or 1, so I've naturally chosen 1 ...  /AA
   */
  while (tries && (MPP_CONFIG3 & 0x40)) tries-- ; if (tries) return 1 ;
//  tries = MPPTRIES ;
//  while (tries && (MPP_CONFIG3 & 0x40)) tries-- ; if (tries) return 1 ;
  if (!(gatos.debug&GATOSQUIET))
    fprintf(stderr,"mpp_wait(): MPP bus timeout in %s\n",str) ;
  timeouts++ ; return 0 ; }

/* ------------------------------------------------------------------------ */
/* Debug and report routines */

#define DUMPREG(NAME)	\
	fprintf(stderr,"%s: %-22s (0x%04X) = 0x%08X\n", \
		ident,#NAME,NAME,tvout_read32(NAME))

void tvout_dumpregs(void) {
  if (!rev) return ;
  DUMPREG(TV_MASTER_CNTL) ;
  DUMPREG(TV_RGB_CNTL) ;
  DUMPREG(TV_CLKOUT_CNTL) ;
  DUMPREG(TV_SYNC_CNTL) ;
  DUMPREG(TV_I2C_CNTL) ;
  DUMPREG(TV_MPP_DATA_CNTL) ;
  DUMPREG(TV_HTOTAL) ;
  DUMPREG(TV_HDISP) ;
  DUMPREG(TV_HSIZE) ;
  DUMPREG(TV_HSTART) ;
  DUMPREG(TV_HCOUNT) ;
  DUMPREG(TV_VTOTAL) ;
  DUMPREG(TV_VDISP) ;
  DUMPREG(TV_VCOUNT) ;
  DUMPREG(TV_FTOTAL) ;
  DUMPREG(TV_FCOUNT) ;
  DUMPREG(TV_FRESTART) ;
  DUMPREG(TV_HRESTART) ;
  DUMPREG(TV_VRESTART) ;
  DUMPREG(TV_SYNC_SIZE) ;
  DUMPREG(TV_TV_PLL_CNTL) ;
  DUMPREG(TV_CRT_PLL_CNTL) ;
  DUMPREG(TV_PLL_CNTL) ;
  DUMPREG(TV_PLL_TEST_CNTL) ;
  DUMPREG(TV_CLOCK_SEL_CNTL) ;
  DUMPREG(TV_FRAME_LOCK_CNTL) ;
  DUMPREG(TV_SYNC_LOCK_CNTL) ;
  DUMPREG(TV_TVO_SYNC_PAT_ACCUM) ;
  DUMPREG(TV_TVO_SYNC_THRESHOLD) ;
  DUMPREG(TV_TVO_SYNC_PAT_EXPECT) ;
  DUMPREG(TV_DELAY_ONE_MAP_A) ;
  DUMPREG(TV_DELAY_ONE_MAP_B) ;
  DUMPREG(TV_DELAY_ZERO_MAP_A) ;
  DUMPREG(TV_DELAY_ZERO_MAP_B) ;
  DUMPREG(TV_TVO_DATA_DELAY_A) ;
  DUMPREG(TV_TVO_DATA_DELAY_B) ;
  DUMPREG(TV_HOST_READ_DATA) ;
  DUMPREG(TV_HOST_WRITE_DATA) ;
  DUMPREG(TV_HOST_RD_WT_CNTL) ;
  DUMPREG(TV_VSCALER_CNTL) ;
  DUMPREG(TV_TIMING_CNTL) ;
  DUMPREG(TV_GAMMA_CNTL) ;
  if (rev == 2) {
    DUMPREG(TV_Y_FALL_CNTL) ;
    DUMPREG(TV_Y_RISE_CNTL) ;
    DUMPREG(TV_Y_SAW_TOOTH_CNTL) ; }
  DUMPREG(TV_MODULATOR_CNTL1) ;
  DUMPREG(TV_MODULATOR_CNTL2) ;
  DUMPREG(TV_PRE_DAC_MUX_CNTL) ;
  DUMPREG(TV_TV_DAC_CNTL) ;
  DUMPREG(TV_CRC_CNTL) ;
  DUMPREG(TV_VIDEO_PORT_SIG) ;
  if (rev == 3) {
    DUMPREG(TV_VBI_20BIT_CNTL) ;
    DUMPREG(TV_VBI_LEVEL_CNTL) ; }
  DUMPREG(TV_UV_ADR) ;
  DUMPREG(TV_FIFO_TEST_CNTL) ;
  /* Report then clear MPP bus timeout stats */
  if (timeouts) fprintf(stderr,"Warning: %d MPP bus timeouts\n",timeouts) ;
  timeouts = 0 ; }

/* ------------------------------------------------------------------------ */
/* Setting up tvout */

//#define RefClock 28.63636
//#define RefClock 28.63
//#define RefClock gatos.refclock
#define VR60Hz 60./1.001
#define VR50Hz 50
#include <unistd.h>  /* for usleep */

struct tvs {
 double RefClock;
 double tvRefClock;
 int format;
 double fieldrate;
 double fsc;
 double hscale,vscale,vshift,hshift;
 int pal;
 int gamma;
 int bright;
 int saturation;
 int flickrem;
 int Csharp;
 int SVIDsharp;
 int LTseries;
 int PROseries;
 int mach64HasDSP;
 int uncertainty;
 int ns_per_tap;
 int datad[8];
} tvp;

/*=====================================================*/

int tvout_start(void) {
  if(!tvout_present()) RETURN(ENODEV);
  tvp.RefClock=gatos.refclock;
  /* set TV format, based on gatos struct */
  if (gatos.format){
    if (VERBOSE) printf("tvout: format set to %d\n",gatos.format);
    tvp.format=gatos.format;
  } else {
    /* none set, default to NTSC */
    /* this shouldn't really happen */
    fprintf(stderr,"tvout: No TV format set, defaulting to NTSC\n");
    tvp.format=1;
  }
  tvp.hscale=tvp.vscale=1.;
  tvp.gamma=70.;
  tvp.bright=100.;
  tvp.saturation=0;
  tvp.flickrem=0;
  tvp.Csharp=1;
  tvp.SVIDsharp=1;
 
  /* Allow MPP access during VBlank and HBlank */
  TVO_CNTL&=~(1<<7);
  /* Make Graphics Chip timing master and reset timing */
  tvout_write32(TV_MASTER_CNTL,0);
  /* Enable CLKOUT, Jon */
  tvout_write32(TV_CLKOUT_CNTL,1);
  /* set SYNC_PUB and SYNC_DRV to default values, Jon */
  tvout_write32(TV_SYNC_CNTL,0x28);
  /* Leave TV_MPP_DATA_CNTL the way it is */
  /* leave macro on defaults (mine are 0xD4)*/
  /* Switch on TV and CRT clock PLL, Jon */
  tvout_write32(TV_PLL_CNTL,tvout_read32(TV_PLL_CNTL)|0x18);
  /* Disable PLL test - set everything to defaults */
  tvout_write32(TV_PLL_TEST_CNTL,0);
  /* Set max frame lock expected strength to maximum (0xf), Jon */
  tvout_write32(TV_FRAME_LOCK_CNTL,0xf);
  /* set default SYNC lock values, cancelling any currently running
     SYNC commands, Jon */
  tvout_write32(TV_SYNC_LOCK_CNTL,0x01000000);
  /* set defaults for HOST_RD_WT_CNTL, Jon */
  tvout_write32(TV_HOST_RD_WT_CNTL,0);
  /* set second modulator delay to 3 cycles and request Y data first, Jon */
  tvout_write32(TV_TIMING_CNTL,0x00090000);
  /* Enable Y/Red, C/Green, Comp/Blue guns of the DAC and DAC dither, Jon */
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0x0000000f);
  /* set defaults for video comapctor and disable, Jon */
  tvout_write32(TV_CRC_CNTL,0);
  /* switch off VBI timing codes (backward compatible), Jon */
  if(rev>2) tvout_write32(TV_VBI_20BIT_CNTL,0);
  /* Set max adderss for UV data in fiforam to 0xBE, Jon */
//  tvout_write32(TV_UV_ADR,0x000000BE); // What should this be?
  tvout_write32(TV_UV_ADR,0x000000BD); // What should this be?
  /* set defaults for FIFO_TEST_CNTL, Jon*/
  if(rev>1) tvout_write32(TV_FIFO_TEST_CNTL,0); 
  if (VERBOSE) { printf("Ready to setmode!\n") ; fflush(stdout) ; }
  RETURN(tvout_setmode()); }

/*=====================================================*/

void tvout_setcont(){
 int gam;
 u32 tmp;
 if (VERBOSE) { printf("Setting Contrast\n") ; fflush(stdout) ; }
 tmp=tvout_read32(TV_GAMMA_CNTL)&0xffff0000;
 gam=tvp.gamma/90.*7;
 if(gam>7) gam=7;
 if(gam<0) gam=0;
 tmp|=gam;
 if(rev>1) tmp|=gam<<8; else tmp|=2<<8;
 if(rev==1) tmp|= (gam<<4)|(gam<<12);
 tvout_write32(TV_GAMMA_CNTL,tmp);
}

/*=====================================================*/

void tvout_set_bright() {
 int bright,min;
 u32 tmp;
 if (VERBOSE) { printf("Setting Brightness\n") ; fflush(stdout) ; }
 tmp=tvout_read32(TV_GAMMA_CNTL)&0xff00ffff;
 if(rev>1) {
   /* also need to set UV_INPUT_PRE_SCALE in TIMING_CONTROL */
   tvout_write32(TV_TIMING_CNTL,(tvout_read32(TV_TIMING_CNTL)&0x00ffffff)|
    (0x20<<24)); /* Set to default -- probably needs adjusting */
   bright=tvp.bright/50.*0x40;
   if(bright>0x7f) bright=0x7f;
   if(bright<0) bright=0;
   tmp|= bright<<16; }
 else {
   bright=tvp.bright/50.*0x20;
   if(bright>0x3f) bright=0x3f;
   if(bright<0) bright=0;
   if(bright>0x1f) {bright-=0x20; min=0; }
   else {bright=0x1f-bright; min=1;}
   tmp|= (bright<<16)|(min<<23); }
 tvout_write32(TV_GAMMA_CNTL,tmp);
}

/*=====================================================*/

void tvout_set_sat() {
  u32 tmp;
  int burst;
  if (VERBOSE) { printf("Setting Saturation\n") ; fflush(stdout) ; }
  tmp=tvout_read32(TV_MODULATOR_CNTL2)&0xffff0000;
  if(tvp.pal) burst=-78;
  else burst=-111;
/* this is a very rough saturation calculation. The values should be limited to
 * a smaller range! */
  burst+=tvp.saturation;
  if(burst>255) burst=255;
  if(burst<-254) burst=-254;
  tmp+= burst&0x1ff;
  tvout_write32(TV_MODULATOR_CNTL2,tmp);
}

/*=====================================================*/

void tvout_set_flick() {
  /* flick=0 for maximum blending == minimum flicker == minimum vertical
   * sharpness  */
  u32 tmp;
  int flick;
 if (VERBOSE) { printf("Setting Flicker removal\n") ; fflush(stdout) ; }
  tmp=tvout_read32(TV_VSCALER_CNTL)&0xff00ffff;
/* here aggain I do a tempory adjustment */
  flick = 128-tvp.flickrem;
  if(flick<0) flick=0;
  if(flick>0x7f) flick=0x7f;
  tmp|=flick<<16;
  tvout_write32(TV_VSCALER_CNTL,tmp);
  tmp=tvout_read32(TV_GAMMA_CNTL)&0x00ffffff;
/* Should also adjust Y_ACCUM_INIT @ GAMMA_CNTL; just setting it to 1:1 */
  if(rev>1) tmp|=0x63<<24;
  else tmp|=1<<24;
  tvout_write32(TV_GAMMA_CNTL,tmp);
}

/*=====================================================*/

void tvout_set_sharp() {
  u32 tmp;
  if (VERBOSE) { printf("Setting Sharpness\n") ; fflush(stdout) ; }
  tmp=tvout_read32(TV_MODULATOR_CNTL1)&0xffffffc0;
/* This is in progress ... */
  tmp|=0xc;
  if(tvp.Csharp!=1) tmp|=0x20;
  if(tvp.Csharp>1) tmp|=1;
  if(tvp.SVIDsharp<1) tmp|=0x10;
  if(tvp.SVIDsharp>1) tmp|=2;
  if(rev>1) {tmp&=0xfdffffff; if(tvp.Csharp>2) tmp|=0x02000000; }
  tvout_write32(TV_MODULATOR_CNTL1,tmp);
}

/*=====================================================*/

int tvout_reset(void) {
  struct impactv_clockinfo clk;
  if (VERBOSE) { printf("Doing TVout RESET\n") ; fflush(stdout) ; }
  if(!tvout_present()) RETURN(ENODEV);
  /* Vref is just a guess */
  if(setclk(55., 0, 0, &clk))
    RETURN(EPERM);
  OVERSCAN_CLR=0; // should have saved these
  OVERSCAN_WLR=0;
  OVERSCAN_WTB=0;
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0x00000008);
  tvout_write32(TV_TV_DAC_CNTL,(tvout_read32(TV_TV_DAC_CNTL)&~9)|8);
  /* may be also disbale something else :PLL? */
  RETURN0; }

/*=====================================================*/

#define PLL_WR_EN               0x02
#define PLL_REF_DIV             0x02
#define PLL_VCLK_CNTL           0x05
#define VCLK_POST_DIV           0x06
#define VCLKx_FB_DIV            0x07
#define PLL_XCLK_CNTL           0x0B

/*=====================================================*/

u8 clock_cntl_r(int index) {
  CLOCK_CNTL1=index<<2 ;
  return CLOCK_CNTL2;
}

/*=====================================================*/

void clock_cntl_w(int index, u8 data) {
  CLOCK_CNTL1=index<<2|PLL_WR_EN ;
  CLOCK_CNTL2=data;
}

/*=====================================================*/

void chipid(void) {
  if (VERBOSE) { printf("Finding chipid\n") ; fflush(stdout) ; }
  tvp.mach64HasDSP = 1;
  if( ati.gx || ati.cx || ati.ct || ati.et ||
      (( ati.vt || ati.gt) && !(ati.doesvtb||ati.doesgtc)))
    tvp.mach64HasDSP = 0;

  tvp.LTseries=0;
  tvp.PROseries=0;
  if(ati.lg) tvp.LTseries=1;
  if(ati.lt_c && !ati.ltm) {
    tvp.LTseries=1; 
    tvp.PROseries=1;}
  if(ati.gt_c) tvp.PROseries=1;

  tvp.RefClock=gatos.refclock;
  tvp.tvRefClock=tvp.RefClock;
}

/*=====================================================*/

void readclk(struct impactv_clockinfo *clk){
  /* determines the current video mode */
  u32 tmp;
  double RefClock;
  int N,M;
  if (VERBOSE) { printf("Reading clocks\n") ; fflush(stdout) ; }
  chipid();
  clk->clksel=CLOCK_CNTL0&3;
  clk->RefDiv=clock_cntl_r(PLL_REF_DIV);
  clk->srcsel=clock_cntl_r(PLL_VCLK_CNTL)&3;
  tmp=(clock_cntl_r(VCLK_POST_DIV)>>(2*clk->clksel))&3;
  clk->FBDiv=clock_cntl_r(VCLKx_FB_DIV+clk->clksel);
  clk->ExtDiv=(clock_cntl_r(PLL_XCLK_CNTL)>>(4+clk->clksel))&1;
  clk->PostDiv=1<<tmp;
  if( tvp.mach64HasDSP && clk->ExtDiv) {
    if(clk->PostDiv==1) clk->PostDiv=3;
    else clk->PostDiv=clk->PostDiv*3/2; }
  if (VERBOSE) { 
    printf("readclk: RefDiv=%02x, FBDiv=%02x, PostDiv=%i, DSP=%s, clksel=%i, srcsel=%i\n",
            clk->RefDiv,clk->FBDiv,clk->PostDiv,
            tvp.mach64HasDSP?"yes":"no",clk->clksel,clk->srcsel) ;
    fflush(stdout) ; }
  RefClock=tvp.RefClock;
  clk->clock=tvp.RefClock*2*clk->FBDiv/(clk->RefDiv*clk->PostDiv);
  if(clk->srcsel==2) { //Clock source is Impact chip
    RefClock=tvp.tvRefClock;
    tmp=tvout_read32(TV_CRT_PLL_CNTL);
    clk->tvM=tmp&0xff;
    if(rev>1) clk->tvN=(tmp>>8)&0x1ff;
    else clk->tvN=(tmp>>8)&0xff;
    clk->tvBytDiv=((tvout_read32(TV_CLOCK_SEL_CNTL)>>2)&3)+1;
    clk->clock=tvp.tvRefClock*clk->tvN/clk->tvM*2/clk->tvBytDiv/clk->PostDiv;
    tmp=tvout_read32(TV_TV_PLL_CNTL);
    M=tmp&0xff;
    if(rev>1) N=tmp&0x1ff;
    else N=tmp&0xff;
    clk->tvclock=tvp.tvRefClock*N/M;
  }
  if (VERBOSE) { printf("readclk: RefClock=%g, clock=%g, bytDiv=%d\n",
                  RefClock,clk->clock,clk->tvBytDiv) ; fflush(stdout) ; }
  tmp=CRTC_H_TOTAL_DISP;
  clk->htotal=((tmp&0x1ff)+1)*8; 
  clk->hactive=(((tmp>>16)&0x1ff)+1)*8;
  clk->hextra=TVO_CNTL&7;
  clk->htotal+=clk->hextra;
  tmp=CRTC_V_TOTAL_DISP;
  clk->vtotal=(tmp&0x7ff)+1;
  clk->vactive=((tmp>>16)&0x7ff)+1;
  tmp=CRTC_H_SYNC_STRT_WID;
  /* note that there is a high bit available and a delay and tvout shift ...*/
  /* (TVO_CNTL>>4)&1 ==1 correction on falling edge of hsync */
//  clk->hstart=((tmp&0xff)+1)*8;
// The next line uses the high bit and delay(whic affects only hstart, not
// htotal
  clk->hstart=(((tmp&0xff)|((tmp>>4)&0x100))+1)*8 + ((tmp>>8)&7);
  clk->hwidth=((tmp>>16)&0x1f)*8;
  clk->hpolarity=(tmp>>21)&1; /* 1 means active low */
  if(TVO_CNTL&0x10) clk->hstart+=clk->hextra;
  else clk->hwidth+=clk->hextra;
  tmp=CRTC_V_SYNC_STRT_WID;
  clk->vstart=(tmp&0x7ff)+1;
  clk->vwidth=(tmp>>16)&0x1f;
  clk->vpolarity=(tmp>>21)&1; /* 1 means active low */

  clk->hrate=clk->clock/clk->htotal*1e3; /* in kHz */
  clk->vrate=clk->hrate/clk->vtotal*1e3;  /* in Hz */
  if (VERBOSE) { printf("readclk: hrate=%g vrate=%g\n",
                  clk->hrate,clk->vrate) ; fflush(stdout) ; }
//Modeline "320x200@70" 12.587   320  352  376  400    200  206  207  225
  if (VERBOSE) { printf("readclk: MODE\nModeline %dx%d@%.0f  %f  %d %d %d %d   %d %d %d %d%s%s%s%s\n",
         clk->hactive,clk->vactive,clk->vrate,clk->clock,
         clk->hactive,clk->hstart,clk->hstart+clk->hwidth,clk->htotal,
         clk->vactive,clk->vstart,clk->vstart+clk->vwidth,clk->vtotal,
         clk->hpolarity?" -hsync":" +hsync",
         clk->vpolarity?" -vsync":" +vsync",
         (CRTC_GEN_CNTL&1)?" DoubleScan":"",
         (CRTC_GEN_CNTL&2)?" Interlaced":"");
                  ; fflush(stdout) ; }
}

/*=====================================================*/

/*Set BytePerPix to 0,2 or 3. 0 is for autoselection */
int setclk(double vfreq, int tvsrc, int BytePerPix, struct impactv_clockinfo *clk) {
  /* sets clocks
     vfreq - TV vertical refresh rate in Hz
     tvsrc - boolean - true if tv is to be clock source
     BytePerPix - bytes per pixel
     clk - clock info structure
     ,Jon */
  double pixclk=0;
  u8 tmp;
  int postdiv[]={1,2,3,4,6,8,12};
  int postdivs[]={0,1,0,2,2,3,3};
  int postdivx[]={0,0,1,0,1,0,1};
  int idx;

  if (VERBOSE) { printf("Setting clocks\n") ; fflush(stdout) ; }
  readclk(clk);
  if(tvsrc) { // going to ImpacTV as clock source
    if(clk->srcsel!=3) {printf("Not in correct mode\n"); return 1;}
    clk->srcsel=2;
    /* barf if CRT and proposed TV refresh rate differ by more than 2 Hz, 
       Jon */
    if(fabs(vfreq-clk->vrate)>2.) {printf("Original frequency too far\n");
                                    return 1;}
    /* vtotal is total number of lines on monitor (displayed and not 
       displayed), CL */
    if(clk->vtotal>1000) {printf("Original vtotal too large\n");
                                    return 1;}
    /* calculate pixclock freq in MHz, Jon */
    pixclk=vfreq*(clk->htotal-clk->hextra)*clk->vtotal/1e6;
    /* barf if it's > 50 MHz */
    if(pixclk>50.) {printf("Original pixclock too high\n"); return 1;}
    /* htotal includes hextra. We do calculation starting from hextra=0 */
    clk->htotal-=clk->hextra;
    clk->hextra=0;
    /* Guess Bytes per pixel from the pixclock frequency */
    if(BytePerPix<=0) {
      if(pixclk<33.33) clk->PostDiv=3;
      else clk->PostDiv=2; }
    else {
      if(BytePerPix==3) clk->PostDiv=3;
      else clk->PostDiv=2; }
    if(clk->PostDiv==2) idx=1; else idx=2;
  } else { // going to RAGE as clock source
    // assuming only the selector and the post divider need to be adjusted
    int i;
    double vf,bestvfdiff,vfdiff;
    clk->srcsel=3;
    vf=tvp.RefClock*2e6*clk->FBDiv/(clk->RefDiv*clk->htotal*clk->vtotal);
    bestvfdiff=fabs(vf/postdiv[0]-vfreq);
    idx=0;
    for(i=1;i<7;i++) {
      vfdiff=fabs(vf/postdiv[i]-vfreq);
      if(vfdiff<bestvfdiff) { bestvfdiff=vfdiff; idx=i; } }
  }
  tmp=clock_cntl_r(PLL_VCLK_CNTL);
  tmp|=4;
  clock_cntl_w(PLL_VCLK_CNTL,tmp); /*Reset the PLL*/
  tmp&=0xfc;
  tmp|=clk->srcsel;
  if(tvsrc) {
    /* make register clock not be the TV clock, Jon */
    tvout_write32(TV_CLOCK_SEL_CNTL, tvout_read32(TV_CLOCK_SEL_CNTL)&~0x10);
    setcrtfreq(pixclk,clk->tvclock,clk->PostDiv,0,clk->htotal,0); }
  else 
    TVO_CNTL=TVO_CNTL&~0x80000007;
  clock_cntl_w(PLL_VCLK_CNTL,tmp);
  tmp=clock_cntl_r(VCLK_POST_DIV);
  tmp&= ~(3<<(2*clk->clksel));
  tmp|= postdivs[idx]<<(2*clk->clksel);
  clock_cntl_w(VCLK_POST_DIV,tmp);
  tmp=clock_cntl_r(PLL_XCLK_CNTL);
  if(postdivx[idx]) tmp|= 1<<(clk->clksel+4);
  else  tmp&= ~(1<<(clk->clksel+4));
  clock_cntl_w(PLL_XCLK_CNTL,tmp);
  usleep(100000);
  if(tvsrc)
    tvout_write32(TV_CLOCK_SEL_CNTL,tvout_read32(TV_CLOCK_SEL_CNTL)|0x10);
  clock_cntl_w(PLL_VCLK_CNTL,clock_cntl_r(PLL_VCLK_CNTL)&~4);/*Set the PLL*/

  readclk(clk);
  return 0;
}

/*=====================================================*/

/* use htotal=0 to prevent calculation of hadd, i.e. keep hadd=0 */
void findNM(double R, u32 *N,u32 *M,u32 htotal, u8 *hadd,int n9bit) {
  /* Determine values of N and M such that they make a fraction close to the 
     value of R, Jon */
  int n,nmax,m,h;
  double diff,mindiff,r;
  if (VERBOSE) {
    printf("FindNM(R=%f,htotal=%d,n9bit=%d,hadd=%d)\n",R,htotal,n9bit,!hadd?0:(int)*hadd) ; 
    fflush(stdout) ; }
  mindiff=1e10;
  if(n9bit) nmax=512; 
  else nmax=256;
  for(n=1;n<nmax;n++)  {
    if(htotal>0) {
      for(h=0;h<8;h++) {
        r=R*(htotal+h)/htotal;
        m=rint(n/r);
        if(m<50) continue;
        if(m==0) m=1;
        if(m>255) m=255;
        diff=fabs(r-1.*n/m);
        if(diff<mindiff) {mindiff=diff; *N=n; *M=m;*hadd=h;} } }
    else {
      m=rint(n/R);
      if(m==0) m=1;
      if(m<50) continue;
      if(m>255) m=255;
      diff=fabs(R-1.*n/m);
      if(diff<mindiff) {mindiff=diff; *N=n; *M=m; } }
  } 
  if (VERBOSE) {
    printf("End of FindNM(M=%d,N=%d,Hadd=%d\n",*M,*N,!hadd?0:(int)*hadd) ; 
    fflush(stdout) ; } }

/*=====================================================*/

void tvout_setgeom(struct impactv_clockinfo p,int htotal, int vtotal) {
  int tmp;
  if (VERBOSE) { printf("Set Geometry\n") ; fflush(stdout) ; }
  tvout_write32(TV_HTOTAL,p.htotal-1);
  tvout_write32(TV_HDISP,p.hactive-1);
  tvout_write32(TV_HSIZE,p.hactive);
  tvout_write32(TV_HSTART,p.hactive+12); /* ????? */
  tvout_write32(TV_VTOTAL,p.vtotal-1);
  tvout_write32(TV_VDISP,p.vactive-1);
  tvout_write32(TV_FTOTAL,1);
  tvout_write32(TV_FRESTART,0);
  tmp=p.hactive+12+tvp.hshift; /* ????? */
  if(tmp<0) tmp=0;
  if(tmp>0x7ff) tmp=0x7ff;
  tvout_write32(TV_HRESTART,tmp);
  tmp=p.vactive+4+tvp.vshift; /* ????? */
  if(tmp<0) tmp=0;
  if(tmp>0x3ff) tmp=0x3ff;
  tvout_write32(TV_VRESTART,tmp);
  tvout_write32(TV_SYNC_SIZE,p.hactive+8);
  /* note that the vsize cannot be adjusted independently of computer monitor*/
  //max p.vtotal for NTSC is 1049
  tmp=p.vtotal*(1<<0xe)*2/vtotal;
  if(tmp<0) tmp=0;
  if(tmp>0xffff) tmp=0xffff;
  tvout_write32(TV_VSCALER_CNTL,(tvout_read32(TV_VSCALER_CNTL)&0xffff0000) 
                      |tmp);
  //max (p.htotal+50) for NTSC is 2729
  // 50 is just to scale picture smaller
  tmp=(p.htotal+50)*(1<<0xc)/htotal;
  /* temporary scalling */
  tmp*=tvp.hscale;
  if(tmp<0) tmp=0;
  if(tmp>0xfff) tmp=0xfff;
  tvout_write32(TV_TIMING_CNTL,(tvout_read32(TV_TIMING_CNTL)&0xffff0000)|tmp);
  if(rev>1) {
  /* set Y_FALL_CNTL,  Y_RISE_CNTL,  Y_SAW_TOOTH_CNTL */
  }
}

/*=====================================================*/

void tvout_setdelay(struct impactv_clockinfo clk) {
  int datad_max=0;
  u32 byte_period,ns_per_phase;
  int mid_data_eye,no_of_phase;
  int i;
  int BytDel;
  u8 del[8];

  if (VERBOSE) { printf("Set Delay\n") ; fflush(stdout) ; }
  for(i=0;i<8;i++) {
  if (VERBOSE) { printf("tvp.datad[%i]=%d\n",i,tvp.datad[i]) ; fflush(stdout) ; }
    if(tvp.datad[i]>datad_max) datad_max=tvp.datad[i]; }
  if (VERBOSE) { printf("datad_max=%i\n",datad_max) ; fflush(stdout) ; }
  no_of_phase=2*clk.tvBytDiv;
  byte_period=1e6/clk.clock/clk.PostDiv;
  ns_per_phase=byte_period/no_of_phase;
  if(byte_period<12500) mid_data_eye=(byte_period-tvp.uncertainty)/3;
  else mid_data_eye=(byte_period-tvp.uncertainty)/2;
  if (VERBOSE) { printf("no_of_phase=%i  byte_period=%i ns_per_phase=%i mid_data_eye=%i\n",no_of_phase,byte_period,ns_per_phase,mid_data_eye) ; fflush(stdout) ; }
  if (VERBOSE) { printf("tvp.uncertainty=%i tvp.ns_per_tap=%i\n",tvp.uncertainty,tvp.ns_per_tap) ; fflush(stdout) ; }
  if(mid_data_eye > datad_max) {
    BytDel=(mid_data_eye-datad_max)/ns_per_phase;
    for(i=0;i<8;i++) {
      del[i]=((mid_data_eye-tvp.datad[i])-(BytDel*ns_per_phase))/tvp.ns_per_tap;
      if(del[i]>47) del[i]=47; }
    if(BytDel!=0) BytDel=no_of_phase-BytDel; }
  else {
    BytDel=(datad_max-mid_data_eye)/ns_per_phase+1;
    for(i=0;i<8;i++) {
      del[i]=(((BytDel*ns_per_phase)+mid_data_eye)-tvp.datad[i])/tvp.ns_per_tap;
      if(del[i]>47) del[i]=47; 
  if (VERBOSE) { printf("del[%i]=%d\n",i,del[i]) ; fflush(stdout) ; } }
//    if(BytDel>(no_of_phase-1)) BytDel=BytDel-no_of_phase; }
    while(BytDel>(no_of_phase-1)) BytDel=BytDel-no_of_phase; }
  if (VERBOSE) { printf("BytDel=%i\n",BytDel) ; fflush(stdout) ; }
  tvout_write32(TV_CLOCK_SEL_CNTL,(tvout_read32(TV_CLOCK_SEL_CNTL)&0xff)|
                          (BytDel<<8));
  tvout_write32(TV_TVO_DATA_DELAY_A,
          (del[3]<<24)|(del[2]<<16)|(del[1]<<8)|(del[0]));
  tvout_write32(TV_TVO_DATA_DELAY_B,
          (del[7]<<24)|(del[6]<<16)|(del[5]<<8)|(del[4]));
}

/*=====================================================*/

int tvout_setmode(void) {
  struct impactv_clockinfo clk;
  int vtotal,htotal;
  int pal;
  int blank,setup;
  u32 tmp;

  /* determine the current video mode */
  readclk(&clk);
  if (VERBOSE) { printf("Set Mode\n") ; fflush(stdout) ; }
  switch(tvp.format) {
    /* generally htotal= fsc/(vtotal*fieldrate/2) */
    /* The impact requires 12*fsc. */
    /* Obtain htotal from bt829 which is based on 4*Fsc */
    case 1: /*NTSC-M*/
    case 2: /*NTSC-J*/
            tvp.fieldrate=VR60Hz;
            tvp.fsc=3.57954545*12;
            htotal=910*3;
            vtotal=525;
            pal=0;
            break;
    case 4: /*PAL-M */
            tvp.fsc=3.575611*12;
            tvp.fieldrate=VR60Hz;
            htotal=909*3;
            vtotal=525;
            pal=1;
            break;
    case 3: /*PAL-B/G, D/K, H, I */
    case 5: /*PAL-N */
            tvp.fsc=4.433618*12;
            tvp.fieldrate=VR50Hz;
            htotal=1035*3;
            vtotal=625;
            pal=1;
            break;
    case 7: /*PAL-Ncomb */
            tvp.fsc=3.582056*12;
            tvp.fieldrate=VR50Hz;
            htotal=917*3;
            vtotal=625;
            pal=1;
            break;
    default: RETURN(EPERM); /* No SECAM */
             break;
  }

  tvp.pal=pal;
  if (VERBOSE) { printf("Set Mode: set tvfreq\n") ; fflush(stdout) ; }
  /* set the TV clock frequency, Jon */
  settvfreq(tvp.fsc,&clk);
  /* put TV DAC into normal working mode (and set PAL bit if applicable)
     Jon */
  tvout_write32(TV_TV_DAC_CNTL,0x13|(pal?0:0x100));
  /* set V_BURST_LEVEL to 62 if PAL, Jon */
  tvout_write32(TV_MODULATOR_CNTL2,
       (tvout_read32(TV_MODULATOR_CNTL2)&0x1ff)|(pal?0x3e0000:0));
  /* Calculate setup and blank levels for TV_MODULATOR_CNTL1, Jon */
  if(pal) {
    blank=55;
    setup=55;
  } else {
    blank=55;
    if(tvp.format==2) setup=55; 
    else setup=66;
  }
  tmp=tvout_read32(TV_MODULATOR_CNTL1);
  /* For revision B and up force the colour ON */
  if(rev>1) tmp&=0xfeffffff;
  tvout_write32(TV_MODULATOR_CNTL1,
       (tmp&0xff80803f)|(pal?0x40:0x00)|(blank<<8)|(setup<<16));
  if (VERBOSE) { printf("Set Mode: set clock\n") ; fflush(stdout) ; }
  if(setclk(tvp.fieldrate,1,0,&clk)) RETURN(EIO);
  tvout_getdelay();
  tvout_setdelay(clk);
  tvout_setgeom(clk,htotal,vtotal);
  if(clk.PostDiv==3) OVERSCAN_CLR=0x00000100;
  else OVERSCAN_CLR=0x00000800;
  OVERSCAN_WLR=0x00010000;
  OVERSCAN_WTB=0x00020000;
  TVO_CNTL|=0x80000000;
  tvout_setcont();
  tvout_set_bright();
  tvout_set_sat();
  tvout_set_flick();
  tvout_set_sharp();
  if (VERBOSE) {sleep(2); gatos_debug1();}
//  RETURN(EIO);
  RETURN(0);}

/*=====================================================*/

void settvfreq(double fsc,struct impactv_clockinfo *clk) {
  /* set the TV Clock frequency, Jon */
  int N,M;
  if (VERBOSE) { printf("Set tvfreq(fsc=%f)...\n",fsc) ; fflush(stdout) ; }
  /* figure out N and M constants, Jon */
  findNM(fsc/tvp.tvRefClock,&N,&M,0,NULL,rev>1);
  /* make the register clock come from the XTALIN pin (not the TV clock)
     Jon */
  tvout_write32(TV_CLOCK_SEL_CNTL,tvout_read32(TV_CLOCK_SEL_CNTL)&~0x20);
  /* write the calculated N and M values to the CRT PLL, Jon */
  tvout_write32(TV_CRT_PLL_CNTL,(N<<8)|(M)); 
  usleep(100000);
  /* make the register clock come from the TV clock, Jon */
  tvout_write32(TV_CLOCK_SEL_CNTL,tvout_read32(TV_CLOCK_SEL_CNTL)|0x20);
  clk->tvclock=tvp.tvRefClock*N/M;
  if (VERBOSE) { printf("Set tvfreq=%f\n",clk->tvclock) ; fflush(stdout) ; }
}

/*=====================================================*/

int setcrtfreq(double pixclk,double tvclock, int PostDiv, int aligner,
              int htotal,int findphase) {
  u8 hextra;
  int N,M;
  int tvBytDiv=floor(180./(pixclk*PostDiv)); /* 180 is my guess, CL */
  if (VERBOSE) { 
    printf(
    "Set crtfreq(pixclk=%f,tvclock=%f,Postdiv=%d,htotal=%d,findphase=%d)...\n",
        pixclk,tvclock,PostDiv,htotal,findphase) ; 
    fflush(stdout) ; }
  if(tvBytDiv>4) tvBytDiv=4;
  if(tvBytDiv<1) tvBytDiv=1;
  if(tvBytDiv>3 && aligner) tvBytDiv=3; /* Maximum for aligner code */
  if(findphase) return 2*tvBytDiv;
  /* Find an appropriate N & M value for the PLL, Jon */
  findNM(pixclk*tvBytDiv*PostDiv / (tvp.tvRefClock*2),
           &N, &M, htotal, &hextra,rev>1);
  /* Set control for CRT clock to be CRT_CLK, 
     Set BYTE clock divider (according to tvBytDiv),
     Unselect test clock for TV and CRT clocks (default),     
     Maybe we should also be setting PIX_CLK_SEL here?       Jon */
  tvout_write32(TV_CLOCK_SEL_CNTL,
          (tvout_read32(TV_CLOCK_SEL_CNTL)&0x731)|((tvBytDiv-1)<<2)|2);
  /* also contains dither modes ...*/
  /* set RGB mode 888 (3 bytes per pixel) or 565 (2 bytes per pixel) , Jon */
  tvout_write32(TV_RGB_CNTL,
          (tvout_read32(TV_RGB_CNTL)&0xc)|(PostDiv==3?1:0));
  /* set the CRT clock frequency, Jon */
  tvout_write32(TV_CRT_PLL_CNTL,(N<<8)|M); 
  if(rev>1)
     tvout_write32(TV_VSCALER_CNTL,(tvout_read32(TV_VSCALER_CNTL)&0xf3ffffff)|
         ((pixclk<.55*tvclock)?0x0d000000:0x09000000));
  else
     tvout_write32(TV_VSCALER_CNTL,(tvout_read32(TV_VSCALER_CNTL)&0xfbffffff)|
         ((pixclk<.55*tvclock)?0x07000000:0x03000000));
  TVO_CNTL=(TVO_CNTL&~0x17)|hextra; 
  return 2*tvBytDiv;
}

/*=====================================================*/

int tvout_monitor(void) {
/* what seems to be neede is 1001 for MON which is 
  bits 12-15 of TV_TV_DAC_CNTL.
  direct access to mpp gives good result:
  m 8 a0
  m 0x18 --> either 33 or 13
  m 8 15  

or setting up: m 8 a0;  MON=1111
  m 0x18
*/ 
  int flag=0;
  if (VERBOSE) { printf("Set tvmonitor \n") ; fflush(stdout) ; }
  if(!tvout_present()) return 0;
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0x00008004);
  tvout_write32(TV_TV_DAC_CNTL,0x0000f113);
  usleep(10000);
  if(tvout_read32(TV_TV_DAC_CNTL)&0x20) flag=1;
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0x00000802);
  usleep(10000);
  if(tvout_read32(TV_TV_DAC_CNTL)&0x20) flag|=2;
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0x00000081);
  usleep(10000);
  if(tvout_read32(TV_TV_DAC_CNTL)&0x20) flag|=4;
  /* enable only composite out for now... */
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0x0000000c);
  tvout_write32(TV_TV_DAC_CNTL,0x00000113);
  return flag;
}

/* ------------------------------------- *
 *     Measure delays                    *
 * ------------------------------------- */

static int first_leading_one_edge[8], first_tailing_zero_edge[8];
static int second_leading_one_edge[8], second_tailing_zero_edge[8];
static int third_leading_one_edge[8], third_tailing_zero_edge[8];
static int No_of_Phases;
static u32 ns_per_phase;
static int first_phase_found, second_phase_found, third_phase_found;

static u32 delay_one_map_b[8];
static u32 delay_one_map_a[8];
static u32 delay_zero_map_b[8];
static u32 delay_zero_map_a[8];
static int offset[8];
static int best_offset;

/*=====================================================*/

int tvout_getdelay(void) {
  u32 htotal,vtotal,hsync,vsync,hoffvpitch,gencntl,daccntl;
  u32 overclr,overlr,overtb,tvo;
  u32 clock_sel_cntl;
  u8  plldiv,pllxdiv,pllsel;
  u8 *buf;
  struct impactv_clockinfo clk,nclk;
  int bit,phase,phase_start,phase_end;
  u32 tap_maska,tap_maskb,tap, temp_tap;
  int uncertainty_found, level_one_found;
  u32 byte_period;
  u32 predac, tvdac;

  if (VERBOSE) { printf("Finding out delays\n") ; fflush(stdout) ; }
  readclk(&clk);
  htotal=CRTC_H_TOTAL_DISP;
  vtotal=CRTC_V_TOTAL_DISP;
  hsync=CRTC_H_SYNC_STRT_WID; 
  vsync=CRTC_V_SYNC_STRT_WID; 
  hoffvpitch=CRTC_OFF_PITCH;
  gencntl=CRTC_GEN_CNTL;
  daccntl=DAC_CNTL;
  overclr=OVERSCAN_CLR;
  overlr=OVERSCAN_WLR;
  overtb=OVERSCAN_WTB;
  tvo=TVO_CNTL;
  clock_sel_cntl=tvout_read32(TV_CLOCK_SEL_CNTL);
  predac=tvout_read32(TV_PRE_DAC_MUX_CNTL);
  tvdac=tvout_read32(TV_TV_DAC_CNTL);
  plldiv=clock_cntl_r(VCLK_POST_DIV);
  pllxdiv=clock_cntl_r(PLL_XCLK_CNTL);
  pllsel=CLOCK_CNTL0&3;
  //init variables;
  if (tvp.LTseries) {
    int data_delay_const[] = {15300,13000,11150,11900,11550,11600,11050,12950};
    for(bit=0;bit<8;bit++) tvp.datad[bit]=data_delay_const[bit];
    tvp.uncertainty = 3250; tvp.ns_per_tap = 325;}
  else if (tvp.PROseries)  {
    int data_delay_const[] = {10400,10000,8500,8000,8200,9300,7800,10200};
    for(bit=0;bit<8;bit++) tvp.datad[bit]=data_delay_const[bit];
    tvp.uncertainty = 2150; tvp.ns_per_tap = 235;}
  else {
    int data_delay_const[] = {14336,13780,11520,11600,11360,13390,11090,14336};
    for(bit=0;bit<8;bit++) tvp.datad[bit]=data_delay_const[bit];
    tvp.uncertainty = 2500; tvp.ns_per_tap = 235;}

  if(setcrtfreq(12.587,clk.tvclock,3,1,400,1)==8)
    { printf("Bad BytDiv setting\n"); return -1;}

  if((buf=(u8 *)malloc(320*200*3))==NULL) 
      {printf("Problem!: cannot allocate buf\n"); exit(1); }
//  for(i=0;i<(320*200*3/4);i++) ((u32 *)buf)[i]=((u32 *)ATIFB)[i];
  memcpy(buf,(char *)ATIFB,320*200*3);
  
  /* setting up  320x200 mode 24bpp , Assume we already are partially in tvout
   * mode */
  /* also turn syncs off */
  CRTC_GEN_CNTL= (gencntl&0xfffff8f0)|0x0000050d;
//  CRTC_GEN_CNTL= (gencntl&0xfffff8f0)|0x00000501; // Syncs no disabled
  /*                                         4=(565) 16bit
                                             5=(888) 24bit
                                             6=(a888) 32bit
                                             2=(322) if DAC_DIRECT=1 
                            DAC_DIREAC(10=ah bit)@DAC_CNTL(31h) */
  /* Powerdown of DACs and 8bit*/
  DAC_CNTL=daccntl|0x00008100;
//  DAC_CNTL=daccntl|0x00000100; //just 8 bits

  /* powerdown TV output */
  tvout_write32(TV_PRE_DAC_MUX_CNTL,0);
  tvout_write32(TV_TV_DAC_CNTL,0x10b);

//Modeline "320x200@70" 12.587   320  352  376  400    200  206  207  225
//+Hsync -Vsync doublescan

  CRTC_H_TOTAL_DISP=0x00270031;
  CRTC_V_TOTAL_DISP=0x018F01C1;
  CRTC_H_SYNC_STRT_WID=0x0003002B;
  CRTC_V_SYNC_STRT_WID=0x0022019B;
  CRTC_OFF_PITCH=0x0A000000;
  /*Set clock */
  clock_cntl_w(VCLK_POST_DIV,plldiv&~(3<<2*pllsel));
  clock_cntl_w(PLL_XCLK_CNTL,pllxdiv|(1<<(4+pllsel)));
  /* may be need to set geometry!*/
  /* 12.587 makes a 70HZ which is visible on my monitor
     10.7892 makes a 60Hz which is recordable on VCR but invisible
     on my  monitor (too small a horizontal rate). */
  No_of_Phases=setcrtfreq(12.587,clk.tvclock,3,1,400,0);
//  No_of_Phases=setcrtfreq(10.7892,clk.tvclock,3,1,400,0);
  readclk(&nclk);
  /* Use an NTSC like mode. It does not matter, it should not be displaying
     on the tv. */
  tvout_setgeom(nclk,910*3,525);
   /* ????? */
 /* These are for playing around ... CL */
 /*
  tvout_write32(TV_HSTART,330);
  tvout_write32(TV_HRESTART,330);
  tvout_write32(TV_VRESTART,403);
  tvout_write32(TV_TVO_SYNC_PAT_EXPECT,0xffffff);
  tvout_write32(TV_TVO_SYNC_PAT_EXPECT,0x0000ff);
  tvout_write32(TV_FRAME_LOCK_CNTL,0x2);
 */
  tvout_write32(TV_HRESTART,100);
  tvout_write32(TV_VRESTART,103);
//  tvout_write32(TV_RGB_CNTL,3);
   /* ????? */
  TVO_CNTL|= 1<<7; /* MPP on VBLANK only */
  TVO_CNTL|= 0x80000000;
  tvout_write32(TV_CLOCK_SEL_CNTL,tvout_read32(TV_CLOCK_SEL_CNTL)|0x10);
  tvout_write32(TV_TVO_DATA_DELAY_A,0);
  tvout_write32(TV_TVO_DATA_DELAY_B,0);
/*  tvout_write32(TV_TVO_DATA_DELAY_A,0x10101010);
  tvout_write32(TV_TVO_DATA_DELAY_B,0x10101010);*/
  OVERSCAN_WLR=0x00010000;
  OVERSCAN_WTB=0x00020000;
  
/*  No_of_Phases=setcrtfreq(12.587,clk.tvclock,3,1,400,0);*/
  byte_period=1e12/(12.587e6*3); /* pixclk is around 12.587 -- good enough*/
  ns_per_phase=byte_period/No_of_Phases;
  readclk(&nclk);
  
  if (VERBOSE) { printf("Finding out delays: doing phase loop\n") ; fflush(stdout) ; }
//  tvout_dumpregs();
  if (VERBOSE) gatos_debug1();
  phase_start=2;
  phase_end=phase_start+No_of_Phases-1;
  first_phase_found=0;
  second_phase_found=0;
  third_phase_found=0;
  best_offset=0;
  for(bit=0;bit<=7;bit++) {
    first_leading_one_edge[bit] = second_leading_one_edge[bit] =
      third_leading_one_edge[bit] = 63;
    first_tailing_zero_edge[bit] = second_tailing_zero_edge[bit] = 
      third_tailing_zero_edge[bit] = 63; }
  
  for(phase=phase_start;phase<=phase_end;phase++) {
    if(phase<No_of_Phases) // set Bytdel@Clock_sel_CNTL to phase
      tvout_write32(TV_CLOCK_SEL_CNTL,(tvout_read32(TV_CLOCK_SEL_CNTL)&0xff)|
                          (phase<<8));
    else // set Bytdel@Clock_sel_CNTL to (phase-No_of_Phases)
      tvout_write32(TV_CLOCK_SEL_CNTL,(tvout_read32(TV_CLOCK_SEL_CNTL)&0xff)|
                          ((phase-No_of_Phases)<<8));
      
    if (VERBOSE) { printf("Phase=%d   P-NoP=%d  phase_start=%d\n",
                   phase,phase-No_of_Phases,phase_start) ; fflush(stdout) ; }
    do_coarse_lck();
    gen_delay_map(0,0);

    if(first_phase_found==0) {
      if(bit_count()>2) first_phase_found=1; }
    else if(second_phase_found==0) {
      second_phase_found=1;
      if(bit_count()<3) {
        if(phase_start==No_of_Phases) break; //get out of phase loop. We bail
                                             // out
        else { 
              phase_start=No_of_Phases;
              phase_end=phase_start+No_of_Phases-1;
              first_phase_found=0;
              second_phase_found=0;
              third_phase_found=0;
              best_offset=0;
              phase=phase_start-1; continue; } } }
    else {
      third_phase_found=1;
      bit_count(); }

    if(first_phase_found==1) 
      for(bit=0;bit<=7;bit++) {
        tap_maska=1;
        tap_maskb=0;
        uncertainty_found=0;
        level_one_found=0;
        if((offset[bit]!=best_offset) || ((delay_one_map_b[bit]| 
            delay_one_map_b[bit]| delay_zero_map_b[bit]| delay_zero_map_a[bit])
            == 0)) {
          gen_delay_map(1,bit);
          }
        for(tap=0;tap<=47;tap++) {
          if(((delay_zero_map_b[bit]&tap_maskb) || (delay_zero_map_a[bit] &
             tap_maska)) && (uncertainty_found == 0)) {
            if(tap==47) temp_tap=63;
            else temp_tap=tap;
            if(third_phase_found == 0) {
              if(second_phase_found==0) 
                first_tailing_zero_edge[bit]=temp_tap;
              else 
                second_tailing_zero_edge[bit]=temp_tap; }
            else third_tailing_zero_edge[bit]=temp_tap; }
          else { 
            if((delay_one_map_b[bit]&tap_maskb) || 
                       (delay_one_map_a[bit] & tap_maska)) {
              uncertainty_found=2;
              if (level_one_found==0) {
                level_one_found=1;
                if(tap!=0) {
                  if(third_phase_found==0) {
                    if(second_phase_found==0) first_leading_one_edge[bit]=tap; 
                    else second_leading_one_edge[bit]=tap; }
		  else third_leading_one_edge[bit]=tap; } } }
	    else {
              uncertainty_found=1;
              if(level_one_found==1) {
                level_one_found=0;
                if(third_phase_found==0) {
                  if(second_phase_found==0) first_leading_one_edge[bit]=63;
                  else second_leading_one_edge[bit]=63; }
                else third_leading_one_edge[bit] = 63; } } } 

	  if(tap_maskb==0) {
            if(tap_maska==0x80000000) { tap_maska=0; tap_maskb=1; }
            else tap_maska=tap_maska<<1; }
          else tap_maskb=tap_maskb<<1; } }

    if((third_phase_found==1)||(phase==phase_end)) {
      calc_param(phase);
      break; /* end phase loop */ } }


  if (VERBOSE) { printf("Finding out delays: reseting mode\n") ; fflush(stdout) ; }
  /*  reseting the mode */
  memcpy((char *)ATIFB,buf,320*200*3);
  free(buf);
  tvout_write32(TV_SYNC_LOCK_CNTL,0x01000000);
  TVO_CNTL=tvo;
  OVERSCAN_CLR=overclr;
  OVERSCAN_WLR=overlr;
  OVERSCAN_WTB=overtb;
  /*Reset clock*/
/*  setcrtfreq(clk.clock,clk.tvclock,clk.PostDiv,0,clk.htotal-clk.hextra,0);*/
  setcrtfreq(clk.clock,clk.tvclock,clk.PostDiv,0,0,0);
  TVO_CNTL=tvo;
  tvout_write32(TV_CLOCK_SEL_CNTL,clock_sel_cntl);
  clock_cntl_w(VCLK_POST_DIV,plldiv);
  clock_cntl_w(PLL_XCLK_CNTL,pllxdiv);
  /* reset CRTC */
  CRTC_H_TOTAL_DISP=htotal;
  CRTC_V_TOTAL_DISP=vtotal;
  CRTC_H_SYNC_STRT_WID=hsync;
  CRTC_V_SYNC_STRT_WID=vsync;
  CRTC_OFF_PITCH=hoffvpitch;
  CRTC_GEN_CNTL=gencntl;
  DAC_CNTL=daccntl;
  tvout_write32(TV_PRE_DAC_MUX_CNTL,predac);
  tvout_write32(TV_TV_DAC_CNTL,tvdac);
  return 0;
}

/*=====================================================*/

int bit_count(void) {
  int count=0,bit;
  for(bit=0;bit<=7;bit++) {
    if((delay_one_map_b[bit]>=0x8000)&& (delay_zero_map_a[bit]>=1)) {
       count++;
       if((offset[bit]<best_offset)||(count==1))
          best_offset=offset[bit]; } }
  if (VERBOSE) { printf("bit_count=%d  Best_offset=%d\n",count,best_offset) ; 
                 fflush(stdout) ; }
  return count;
}

/*=====================================================*/

static int valid_pairs;
static int uncertainty;
static int phase_shift_accum;
static int avg_count;
static int const_accum;

/*=====================================================*/

/* Using the constants lets find DATA_DELAY_CONST, UNCERTAINTY NSPER_TAP */
void calc_param(int phase) {
  int bit;
  if (VERBOSE) { printf("Calc Param\n") ; fflush(stdout) ; }
  valid_pairs=0;
  uncertainty=0;
  phase_shift_accum=0;
  for(bit=0;bit<=7;bit++) {
    phase_accum(first_leading_one_edge[bit],second_leading_one_edge[bit]);
    phase_accum(second_leading_one_edge[bit],third_leading_one_edge[bit]);
    phase_accum(first_tailing_zero_edge[bit],second_tailing_zero_edge[bit]);
    phase_accum(second_tailing_zero_edge[bit],third_tailing_zero_edge[bit]);
    find_uncert(first_tailing_zero_edge[bit],first_leading_one_edge[bit]);
    find_uncert(second_tailing_zero_edge[bit],second_leading_one_edge[bit]);
    find_uncert(third_tailing_zero_edge[bit],third_leading_one_edge[bit]); }
  /* exit in case of problems */
  if (VERBOSE) { 
    printf(
       "Calc Param:\n phase_shift_accum=%i, uncertainty=%i, valid_pairs=%i\n",
       phase_shift_accum,uncertainty,valid_pairs ) ; fflush(stdout) ; }
  if((phase_shift_accum==0)||(uncertainty==0)||(valid_pairs==0)) return;
  tvp.uncertainty=uncertainty;
  tvp.ns_per_tap=ns_per_phase*valid_pairs/phase_shift_accum; 
  tvp.uncertainty*=tvp.ns_per_tap;
  if (VERBOSE) { printf("Calc Param OK:\n uncertainty=%d, ns_per_tap=%d\n",
         tvp.uncertainty, tvp.ns_per_tap) ; fflush(stdout) ; }
  for(bit=0;bit<=7;bit++) {
    avg_count=0;
    const_accum=0;
    find_accum(first_leading_one_edge[bit],2,phase);
    find_accum(second_leading_one_edge[bit],1,phase);
    find_accum(third_leading_one_edge[bit],0,phase);
    if(avg_count>0) {
      tvp.datad[bit]=const_accum/avg_count; }
      if (VERBOSE) { printf("  tvp.datad[%i]=%d\n",bit, tvp.datad[bit]);
          fflush(stdout) ; } }
}  

/*=====================================================*/

void phase_accum(int a,int b) {      
  if((a<63)&&(b<63)) {
    phase_shift_accum+= b-a;
    valid_pairs++; } }

/*=====================================================*/

void find_uncert(int a, int b) {
  int new_uncertainty;
  if((a<63)&&(b<63)) {
    new_uncertainty = b-a;
    if(new_uncertainty>uncertainty) 
      uncertainty=new_uncertainty; } } 

/*=====================================================*/

void find_accum(int a, int s, int phase) {
  if(a<63) {
    avg_count++;
    const_accum+=((phase-s)*ns_per_phase)-(a*tvp.ns_per_tap); } }

/*=====================================================*/

#define MAX_SYNC_LOOP 10000

void do_sync_com(void) {
  /* Set trigger */
  u32 tmp;
  int i=0;
  if (VERBOSE) { printf("Doing Sync command\n") ; fflush(stdout) ; }
  tvout_write32(TV_SYNC_LOCK_CNTL,tvout_read32(TV_SYNC_LOCK_CNTL)|(1<<21));
  /* wait for completion */
  while((tmp=tvout_read32(TV_SYNC_LOCK_CNTL))&0x00100000 && i++<MAX_SYNC_LOOP)
    // usleep(10000); 
    // printf("do_sync_com: tmp=%x\n",tmp);
    ;
  if(VERBOSE) {
      // usleep(100000);
      printf("Took %d iter %s\n",i,(i==(MAX_SYNC_LOOP+1))?"MAXED":"") ; 
      fflush(stdout) ; }
  /* reset trigger */
  tvout_write32(TV_SYNC_LOCK_CNTL,tvout_read32(TV_SYNC_LOCK_CNTL)&~(1<<21));
  if (VERBOSE) { printf("Doing Sync command -DONE\n") ; fflush(stdout) ; } }

/*=====================================================*/

#define MPPAddData_ignore 0x0000 //nothing...
//#define MPPAddData_ignore 0xc000 //both
//#define MPPAddData_ignore 0x8000 //just Data
//#define MPPAddData_ignore 0x4000 //just Add

void do_coarse_lck(void) {
  int i;
  if (VERBOSE) { printf("Doing Coarse lock command\n") ; fflush(stdout) ; }
  memset((char *)ATIFB,0,320*200*3); /* blank display */
  OVERSCAN_CLR=0xffffffff;
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((3*328-2)<<16)|(3*328-2));
  tvout_write32(TV_TVO_SYNC_THRESHOLD,((3*328-2)<<16)|(3*328+5));
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((328-2)<<16)|(3*328+5));
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((3)<<16)|(3*328+5));
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((2*808-5)<<16)|(3*808+9));
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((20)<<16)|(4*328));
  /* We need Detet_en=0 for all sync commands */
  tvout_write32(TV_SYNC_LOCK_CNTL,0|MPPAddData_ignore);
  do_sync_com();
  // if (VERBOSE) sleep(2);

  /* prepare gen_delay_map */
  /* Set display to alternating RGB: FF00FF, 00FF00 */
  for(i=0;i<(320*200*3);i+=2) ATIFB[i]=0xff;
  OVERSCAN_CLR=0;
  tvout_write32(TV_TVO_SYNC_THRESHOLD,((3*320/2)<<16)|(3*320/2));
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((3*320/12)<<16)|(4*320));
//  tvout_write32(TV_TVO_SYNC_THRESHOLD,((3*320/20)<<16)|(3*320/2));
}

/*=====================================================*/

void gen_delay_map(int manual,int bitn){
  int bit,bitmin=0,bitmax=7;
  if (VERBOSE) { printf("Generating the delay Map: manual=%i, bitn=%i\n",
                   manual,bitn) ; fflush(stdout) ; }
  if(manual) {
    tvout_write32(TV_SYNC_LOCK_CNTL, 0x04010000|(best_offset<<27)|
                                     MPPAddData_ignore);
    bitmin=bitmax=bitn;
  } else {
    tvout_write32(TV_SYNC_LOCK_CNTL, 0x02010000|MPPAddData_ignore);
  }
  for(bit=bitmin;bit<=bitmax;bit++) {
    tvout_write32(TV_SYNC_LOCK_CNTL,(tvout_read32(TV_SYNC_LOCK_CNTL)&~7)|bit);
    do_sync_com();
    delay_one_map_b[bit]=tvout_read32(TV_DELAY_ONE_MAP_B)&0xffff;
    delay_one_map_a[bit]=tvout_read32(TV_DELAY_ONE_MAP_A);
    delay_zero_map_b[bit]=tvout_read32(TV_DELAY_ZERO_MAP_B)&0xffff;
    delay_zero_map_a[bit]=tvout_read32(TV_DELAY_ZERO_MAP_A); 
    /*if(!manual) */
    offset[bit]=(tvout_read32(TV_TVO_SYNC_PAT_ACCUM)>>11)&0x1f ;
    if (VERBOSE) { 
           printf(" map1b[%d]=%04X 1a=%08X 0b=%04X 0a=%08X offset=%02X\n",
           bit,
           delay_one_map_b[bit],delay_one_map_a[bit],delay_zero_map_b[bit],
           delay_zero_map_a[bit],offset[bit]) ; fflush(stdout); }
  }
  if (VERBOSE) { printf("Generating the delay Map ! DONE!\n") ; fflush(stdout) ; }
}

/*=====================================================*/

int tvout_host_fifo_write(u16 address, u32 data){
  /* 
     Purpose:  Writes 28 least significant bits of data to FIFO
               address specified by 9 LSBs of address
     Returns:  0 on success
               >0 and moans to stderr in case of problems
     Notes:    Makes no check as whether anything else is reading or
               writing
   */

  u32 value;
  const int ntries=100;
  int itry,xsuccess;

  /* check ranges of input parameters */
  /* address */
  if (address & 0xFE00){
    fprintf(stderr,"tvout_host_fifo_write: address %x out of range\n",address);
    return(1);
  }
  /* data */
  if (data & 0xF0000000){
    fprintf(stderr,"tvout_host_fifo_write: data %x out of range\n",data);
    return(1);
  }

  /* write the data to HOST_WRITE_DATA */
  tvout_write32(TV_HOST_WRITE_DATA,data);

  /* write the destination addess into HOST_RD_WT_CNTL:HOST_ADR */
  value = (u32) address;
  tvout_write32(TV_HOST_RD_WT_CNTL,value);

  /* set the HOST_RD_WT_CNTL:HOST_FIFO_WRITE flag */
  value |= 1<<14;
  tvout_write32(TV_HOST_RD_WT_CNTL,value);

  /* wait for the HOST_RD_WT_CNTL:HOST_FIFO_WT_ACK flag to come up */
  for (itry=0,xsuccess=0;(itry<ntries)&&(!xsuccess);itry++){
    xsuccess = ((tvout_read32(TV_HOST_RD_WT_CNTL)&1<<15) != 0);
  }

  /* zero the HOST_RD_WT_CNTL register */
  value=0;
  tvout_write32(TV_HOST_RD_WT_CNTL,value);

  /* did the ack flag come up? */
  if (!xsuccess){
    fprintf(stderr,"tvout_host_fifo_write:  timeout\n");
    return(1);
  }

  return(0);
}

/*=====================================================*/

u32 tvout_host_fifo_read(u16 address){
  /* 
     Purpose:  Reads FIFO address specified by 9 LSBs of address
     Returns:  data on success
               0 and moans to stderr in case of problems
     Notes:    Makes no check as whether anything else is reading or
               writing
   */

  u32 value;
  const int ntries=100;
  int itry,xsuccess;

  /* check ranges of input parameters */
  /* address */
  if (address & 0xFE00){
    fprintf(stderr,"tvout_host_fifo_read: address %x out of range\n",address);
    return(0);
  }

  /* write the source addess into HOST_RD_WT_CNTL:HOST_ADR */
  value = (u32) address;
  tvout_write32(TV_HOST_RD_WT_CNTL,value);

  /* set the HOST_RD_WT_CNTL:HOST_FIFO_RD flag */
  value |= 1<<12;
  tvout_write32(TV_HOST_RD_WT_CNTL,value);

  /* wait for the HOST_RD_WT_CNTL:HOST_FIFO_RD_ACK flag to come up */
  for (itry=0,xsuccess=0;(itry<ntries)&&(!xsuccess);itry++){
    xsuccess = ((tvout_read32(TV_HOST_RD_WT_CNTL)&1<<13) > 0);
  }

  /* zero the HOST_RD_WT_CNTL register */
  value=0;
  tvout_write32(TV_HOST_RD_WT_CNTL,value);

  /* did the ack flag come up? */
  if (!xsuccess){
    fprintf(stderr,"tvout_host_fifo_read:  timeout reading 0x%x\n",address);
    return(0);
  }

  return(tvout_read32(TV_HOST_READ_DATA));
}

/*=====================================================*/
